#!/usr/local/bin/perl

package AbsenceDB;

use FileHandle;
use Fcntl ':flock'; # import LOCK_* constants
use Carp;

use AbsenceConfig;
use AbsenceDate;
use AbsenceLog;

BEGIN {
	$DB_FILE		= AbsenceConfig::fetch('database_file');
	$MODTIME_FILE	= AbsenceConfig::fetch('modtime_db_file');
	%PERSON_KEYS = (
		'name'		=> '',
		'id'		=> '',
		'email'		=> '',
		'group'		=> '',
		'pass'		=> '',
	);
	@RES_KEYS		= qw(id person_id start end type desc misc);
	%RESERVATION_KEYS = (
		'id'		=> '',		# unique number
		'person_id'	=> '',		# unique number
		'start'		=> '',		# date. array-ref in form [d,m,y]
		'end'		=> '',		# date. array-ref in form [d,m,y]
		'type'		=> '',		# single word "vacation", "busy", "travel"
		'desc'		=> '',		# arbitrary text
		'misc'		=> '',		# more arbitrary text
	);
	%GROUP_KEYS = (
		'id'		=> '',
		'name'		=> '',
		'pass'		=> '',
	);
	%PEOPLE			= ();
	%RESERVATIONS	= ();
	%GROUPS			= ();
	$GROUP_COUNT	= 0;
	$FH				= '';
	$INIT_RUN		= 0;
	$DEBUG			= 0;
	$LOGFILE		= '/tmp/absence.log';
	$MT_LOADED		= 0;
	$VERSION		= '1.1';
}

# returns an array of person_ids sorted by name
sub getPeople
{
	my $group_id = shift;

	my @out;

	if (!$INIT_RUN) { init(); }

	if ($group_id == 0) {
		return sort {$PEOPLE{$a}->{name} cmp $PEOPLE{$b}->{name}} keys(%PEOPLE);
	}

	if (!exists($GROUPS{$group_id})) {
		die "getPeople: gid [$group_id] does not exist";
	}

	foreach $pid (keys(%PEOPLE)) {
		if ($PEOPLE{$pid}->{group} == $group_id) {
			push(@out, $pid);
		}
	}

	#@out = sort {$PEOPLE{$a}->{name} cmp $PEOPLE{$b}->{name}} keys(%PEOPLE);

	return sort {$PEOPLE{$a}->{name} cmp $PEOPLE{$b}->{name}} @out;
}

sub getPerson
{
	my $id = shift;

	if (!$INIT_RUN) { init(); }

	if (!exists($PEOPLE{$id})) {
		print "no person with ID=[$id]\n";
		return undef;
	}
	
	return wantarray
		? ($PEOPLE{$id}->{name}, $PEOPLE{$id}->{email}, $PEOPLE{$id}->{group})
		: $PEOPLE{$id}->{name};
}

sub getReservations
{
	my $person_id = shift;

	if (!$INIT_RUN) { init(); }
	my @out;

	$DEBUG && abslog("getReservations for id=[$person_id]");

	foreach my $res_id (keys(%RESERVATIONS)) {
		$DEBUG && abslog("  res_id = [$res_id]");
		if ($person_id == $RESERVATIONS{$res_id}->{person_id}) {
			$DEBUG && abslog("  saving...");
			push(@out, $res_id);
		}
	}

	@out;
}

sub getMonthReservations
{
	my ($person_id, $month, $year) = @_;

	if (!$INIT_RUN) { init(); }
	my @out;

	$DEBUG && abslog("getMonthReservations: p_id = $person_id, m=$month, y=$year");
	foreach my $res_id (keys(%RESERVATIONS)) {
		if ($person_id != $RESERVATIONS{$res_id}->{person_id}) { next; }
		my ($id, $sd, $sm, $sy, $ed, $em, $ey, $desc);
		my $s = $RESERVATIONS{$res_id}->{start};
		my $e = $RESERVATIONS{$res_id}->{end};
		($sd, $sm, $sy) = @$s;
		($ed, $em, $ey) = @$e;
		my $bot = $sm + ($sy - 2000) * 12;
		my $top = $em + ($ey - 2000) * 12;

		my $this = $month + ($year - 2000) * 12;
		if (($this >= $bot) && ($this <= $top)) {
			$s = (($sm == $month) && ($sy == $year)) ? $sd : 1;
			$e = (($em == $month) && ($ey == $year))
				? $ed
				: AbsenceDate::daysInMonth($month, $year);
			push(@out, [
				#$RESERVATIONS{$res_id}->{id},
				$res_id,
				$s, $e,
				$RESERVATIONS{$res_id}->{type},
				$RESERVATIONS{$res_id}->{desc},
			]);
		}
	}

	@out;
}

sub _getMonthReservations
{
	my ($person_id, $month, $year) = @_;

	if (!$INIT_RUN) { init(); }
	my @out;

	$DEBUG && abslog("getMonthReservations: p_id = $person_id, m=$month, y=$year");
	foreach my $res (keys(%RESERVATIONS)) {
		if ($person_id != $RESERVATIONS{$res}->{person_id}) { next; }
		#my ($id, $sd, $sm, $sy, $ed, $em, $ey, $desc) = @$absence;
		my ($id, $sd, $sm, $sy, $ed, $em, $ey, $desc);
		my $s = $RESERVATIONS{$res}->{start};
		my $e = $RESERVATIONS{$res}->{end};
		#($sd, $sm, $sy) = split(/\./, $s);
		#($ed, $em, $ey) = split(/\./, $e);
		($sd, $sm, $sy) = @$s;
		($ed, $em, $ey) = @$e;
		if (($sm == $month) && ($sy == $year)) {
			# if absence starts in month and question...
			if ($em == $month) {
				# ... and also end in same month
				push(@out, [$RESERVATIONS{$res}->{id}, $sd, $ed, $RESERVATIONS{$res}->{type}, $RESERVATIONS{$res}->{desc}]);
			} else {
				# ... or ends in some other month
				push(@out, [$RESERVATIONS{$res}->{id}, $sd, AbsenceDate::daysInMonth($month), $RESERVATIONS{$res}->{type}, $RESERVATIONS{$res}->{desc}]);
			}
		} elsif (($em == $month) && ($ey == $year)) {
			# if absence starts in previous month but ends in current month
				push(@out, [$RESERVATIONS{$res}->{id}, 1, $ed, $RESERVATIONS{$res}->{type}, $RESERVATIONS{$res}->{desc}]);
		}
	}

	@out;
}

sub getReservation
{
	my $rid = shift;

	if (!$INIT_RUN) { init(); }
	if (!exists($RESERVATIONS{$rid})) {
		print "no reservation with ID=[$rid]\n";
		return;
	}
	return $RESERVATIONS{$rid};
}

sub nextPersonId
{
	#if (!$INIT_RUN) { init(); }
	my $highest = 0;
	foreach my $id (keys(%PEOPLE)) {
		if ($id > $highest) {
			$highest = $id;
		}
	}

	++$highest;
}

sub nextGroupId
{
	#if (!$INIT_RUN) { init(); }
	my $highest = 0;
	foreach my $id (keys(%GROUPS)) {
		if ($id > $highest) {
			$highest = $id;
		}
	}

	++$highest;
}

sub nextReservationId
{
	my $highest = 0;
	foreach my $id (keys(%RESERVATIONS)) {
		if ($id > $highest) {
			$highest = $id;
		}
	}

	++$highest;
}

sub addGroup
{
	my ($gid, $name, $pass) = @_;

	$INIT_RUN = 1;	# hack

	$DEBUG && abslog("addGroup: start: name=[$name], pass=[$pass]");

	openFile();
	lockFile();
	readFile();

	my $group_id;

	if ($gid eq 'new') {
		foreach my $group (getGroups()) {
			if ($name eq $GROUPS{$group}->{name}) {
				unlockFile();
				closeFile();
				abslog("addGroup: dup, bailing.");
				return 'duplicate';
			}
		}
		$group_id = nextGroupId();
	} else {
		$group_id = $gid;
		if (!exists($GROUPS{$group_id})) {
			die "error: attempt to modify non-existent group [$group_id]";
		}
	}

	$DEBUG && abslog("addGroup: group-id = [$group_id].");
	abslog("addGroup: name=[$name], pass=[$pass]");
	$GROUPS{$group_id} = {
		id		=> $group_id,
		name	=> $name,
		pass	=> $pass,
	};
	writeFile();
	unlockFile();
	closeFile();

	return 'ok';
}

sub deleteGroup
{
	my $group_id = shift;

	$INIT_RUN = 1;	# hack

	$DEBUG && abslog("deleteGroup: start: name=[$name], pass=[$pass]");

	openFile();
	lockFile();
	readFile();

	exists($GROUPS{$group_id}) || die "deleteGroup: gid [$group_id] does not exist";

	# first, get rid of the members
	foreach my $pid (keys(%PEOPLE)) {
		if ($PEOPLE{$pid}->{group} == $group_id) {
			delete($PEOPLE{$pid});
		}
	}

	# now get rid of the group
	delete($GROUPS{$group_id});
	
	$DEBUG && abslog("deleteGroup: group-id = [$group_id].");
	abslog("deleteGroup: name=[$name], pass=[$pass]");

	writeFile();
	unlockFile();
	closeFile();

	return 'ok';
}

sub getGroups
{
	if (!$INIT_RUN) { init(); }

	my @out;
	foreach my $group_id (keys(%GROUPS)) {
		push(@out, $group_id);
	}

	@out;
}

sub getGroupInfo
{
	my $gid = shift;
	
	if (!$INIT_RUN) { init(); }

	if (!exists($GROUPS{$gid})) {
		print "no group with ID=[$gid]\n";
		return undef;
	}
	
	return wantarray
		? ($GROUPS{$gid}->{name}, $GROUPS{$gid}->{pass})
		: $GROUPS{$gid}->{name};
}

sub getGroup
{
	my $pid = shift;

	if (!$INIT_RUN) { init(); }

	exists($PEOPLE{$pid}) || die "getGroup: pid=[$pid] does not exist";

	return $PEOPLE{$pid}->{group};
}

sub groupCount
{
	$GROUP_COUNT;
}

sub addPerson
{
	my ($pid, $name, $email, $group) = @_;

	$INIT_RUN = 1;	# hack

	$DEBUG && abslog(["addPerson: start: pid=[$pid], name=[$name]", "email=[$email]", "group=[$group]"]);

	openFile();
	lockFile();
	readFile();

	my $person_id;

	if ($pid eq 'new') {
		foreach my $person (getPeople()) {
			if ($name eq $PEOPLE{$person}->{name}) {
				unlockFile();
				closeFile();
				abslog("addPerson: dup, bailing.");
				return 'duplicate';
			}
		}
		$person_id = nextPersonId();
	} else {
		$person_id = $pid;
		if (!exists($PEOPLE{$person_id})) {
			die "error: attempt to modify a non-existent person [$person_id]";
		}
	}

	$DEBUG && abslog("addPerson: person-id = [$person_id].");
	abslog(["addPerson: name=[$name]", "email=[$email]", "id = [$person_id]",
		"group = [$group]"]);
	$PEOPLE{$person_id} = {
		id		=> $person_id,
		name	=> $name,
		email	=> $email,
		group	=> $group,
	};
	writeFile();
	unlockFile();
	closeFile();

	return 'ok';
}

sub deletePerson
{
	my $pid = shift;

	abslog("deletePerson: deleting id=[$pid]");

	openFile();
	lockFile();
	readFile();

	if (!exists($PEOPLE{$pid})) {
		abslog("deletePerson: id=[$pid] not found");
		unlockFile();
		closeFile();
		return 'bad-id';
	}

	# remove all reservations of this person
	foreach my $res_id (keys(%RESERVATIONS)) {
		if ($RESERVATIONS{$res_id}->{person_id} == $pid) {
			abslog("deletePerson: deleting res=[$res_id]");
			delete($RESERVATIONS{$res_id});
		}
	}

	delete($PEOPLE{$pid});

	writeFile();
	unlockFile();
	closeFile();

	return 'ok';
}

#----------------------------------------------------------------------
# addReservation()
#
# adds or modifies a reservation for a person.
# $start and $end are dates in the form dd.mm.yyyy
#----------------------------------------------------------------------

sub addReservation
{
	my ($res_id, $person_id, $start, $end, $type, $desc) = @_;

	my $s = [split(/\./, $start)];
	my $e = [split(/\./, $end)];

	# ROBBO!
	#if ($s->[0] > AbsenceMonth::daysInMonth($s->[1], $s->[2]))
	#}

	$DEBUG && abslog("addReservation: rid=$res_id, pid=$person_id, start=$start, end=$end, type=$type, desc=[$desc]");

	$INIT_RUN = 1;	# hack
	openFile();
	lockFile();
	readFile();

	my $new = 0;

	if ($res_id eq 'new') {
		$new = 1;
		$res_id = nextReservationId();
		$DEBUG && abslog("addReservation: new res-id=[$res_id]");
	} else {
		if (!exists($RESERVATIONS{$res_id})) {
			die "error: attempt to modify non-existent reservation [$res_id]";
		}
		# a reservation is being modified, I need the old start/end dates
		$old_s = $RESERVATIONS{$res_id}->{start};
		$old_e = $RESERVATIONS{$res_id}->{end};
	}

	if (($s->[0] > AbsenceDate::daysInMonth($s->[1])) ||
		($e->[0] > AbsenceDate::daysInMonth($e->[1])))
	{
		return 'badday';
	}

	abslog(
		[
			"addReservation: res=[$res_id]",
			"person-id = [$person_id]",
			"start = [$start]",
			"end = [$end]",
			"type = [$type]",
			"desc = [$desc]",
		]
	);

	my $sjd = AbsenceDate::julianDay(@$s);
	my $ejd = AbsenceDate::julianDay(@$e);
	if ($sjd > $ejd) {
		abslog("addReservation: impossible");
		unlockFile();
		closeFile();
		return 'impossible';
	}

	if (resConflict($res_id, $person_id, $s, $e)) {
		abslog("addReservation: conflict");
		unlockFile();
		closeFile();
		return 'conflict';
	}

	_addReservation($res_id, $person_id, $s, $e, $type, $desc);
	writeFile();

	my $gid = getGroup($person_id);

	if (!$new) {
		updateModificationTimes($gid, $old_s, $old_e);
	}
	updateModificationTimes($gid, $s, $e);

	$DEBUG && abslog("addReservation: success");
	unlockFile();
	closeFile();

	return 'ok';
}

#----------------------------------------------------------------------
# updateModificationTimes()
#
# parameters:
#	$s, $e references to dates in the form:
#		[<day>, <month>, <year>]
#----------------------------------------------------------------------

sub updateModificationTimes
{
	# ROBBO TODO

	my ($gid, $s, $e) = @_;

	$DEBUG && abslog("updateModificationTimes: gid: [$gid], s: [$s->[0]/$s->[1]/$s->[2]], e: [$e->[0]/$e->[1]/$e->[2]]");

	my $openstr = (-e $MODTIME_FILE) ? "+<$MODTIME_FILE" : ">$MODTIME_FILE";
	my $fh = FileHandle->new($openstr);
	defined($fh) || die "open on $MODTIME_FILE failed";

	# first read values
	readModificationTimes($fh);

	# modify in-memory hash
	my $month = $s->[1];
	my $year = $s->[2];

	my $end = $e->[2] * 12 + $e->[1];

	while(($year*12+$month) <= $end) {
		$DEBUG && abslog("updModTime: updating y=$year, m=$month");
		$MT{$gid}->{$year}->{$month} = time();
		if ($month == 12) {
			$year++;
			$month = 1;
		} else {
			$month++;
		}
	}

	# rewind file-pointer
	seek($fh, 0, 0) || die "seek";
	truncate($fh, 0) || die "truncate";

	# ... and write new centents
	foreach $gid (keys(%MT)) {
		foreach $year (keys(%{$MT{$gid}})) {
			foreach $month (keys(%{$MT{$gid}->{$year}})) {
				print $fh "${gid}-$month/$year:$MT{$gid}->{$year}->{$month}\n";
			}
		}
	}

	$fh->close;
	$MT_LOADED		= 1;
}

sub readModificationTimes
{
	my $fh = shift;

	my $do_close = 0;

	my $openstr;

	if (!defined($fh)) {
		if (!-e $MODTIME_FILE) { $MT_LOADED = 1; return; }
		$fh = FileHandle->new("<$MODTIME_FILE");
		defined($fh) || die "open on $MODTIME_FILE failed";
		$do_close = 1;
	}

	while(<$fh>) {
		if (m!^(\d+)-(\d+)/(\d+):(\d+)$!) {
			my ($gid, $month, $year, $time) = ($1, $2, $3);
			$MT{$gid}->{$year}->{$month} = $time;
		}
	}

	if ($do_close) {
		$fh->close;
	}
	$MT_LOADED		= 1;
}

sub getModificationTime
{
	my ($gid, $month, $year) = @_;

	$DEBUG && abslog("getModificationTime: lookup: gid=$gid, m=$month, y=$year");

	if (!$MT_LOADED) {
		readModificationTimes();
	}
	if (exists($MT{$gid}->{$year}) && exists($MT{$gid}->{$year}->{$month})) {
		$DEBUG && abslog("getModificationTime: returning [$MT{$gid}->{$year}->{$month}]");
		return $MT{$gid}->{$year}->{$month};
	}

	$DEBUG && abslog("getModificationTime: returning [undef]");
	return undef;
}

sub deleteReservation
{
	my $res_id = shift;

	if (!$INIT_RUN) { init(); }

	abslog("deleteReservation: res-id=[$res_id]");

	# check if there is such a reservation
	if (!exists($RESERVATIONS{$res_id})) {
		abslog("deleteReservation: ID not found");
		#logMsg("tried to delete non-existent reservation_id [$res_id]");
		return 'internalerror';
	}

	logMsg("delete reservation [$res_id]");
	openFile();
	lockFile();
	readFile();

	if (!exists($RESERVATIONS{$res_id})) {
		# someone else got there first
		abslog("deleteReservation: ID disappeared (overlapping ops)");
		unlockFile();
		closeFile();
		return 'conflict';
	}

	my ($start, $end) =
		($RESERVATIONS{$res_id}->{start}, $RESERVATIONS{$res_id}->{end});

	my $pid = $RESERVATIONS{$res_id}->{person_id};
	delete($RESERVATIONS{$res_id});

	writeFile();
	unlockFile();
	closeFile();

	my $gid = getGroup($pid);
	abslog("deleteReservation: success");
	updateModificationTimes($gid, $start, $end);

	return 'ok';
}

sub resConflict
{
	my ($res_id, $person_id, $start, $end) = @_;
	
	my $sjd = AbsenceDate::julianDay(@$start);
	my $ejd = AbsenceDate::julianDay(@$end);

	$DEBUG && abslog("resConflict: sjd = $sjd, ejd = $ejd");

	foreach my $rid (getReservations($person_id)) {
		$DEBUG && abslog("loop: [$rid] $RESERVATIONS{$rid}->{sjd}, $RESERVATIONS{$rid}->{ejd}");
		if ($res_id == $rid) { next; }	# avoid checking the same res
		my ($rs,$re) = ($RESERVATIONS{$rid}->{sjd}, $RESERVATIONS{$rid}->{ejd});

		if ((($rs >= $sjd) && ($rs <= $ejd)) ||
			(($re >= $sjd) && ($re <= $ejd)))
		{
			return 1;
		}
	}

	return 0;
}

sub _addReservation
{
	my ($res_id, $person_id, $start, $end, $type, $desc) = @_;

	#my $next_id = nextReservationId();
	$DEBUG && abslog("_addReservation: res_id = [$res_id].");

	$RESERVATIONS{$res_id} = {
		id			=> $res_id,
		person_id	=> $person_id,
		start		=> $start,
		end			=> $end,
		type		=> $type,
		desc		=> $desc
	};
}

sub init
{
	$DEBUG && abslog("init: read DB");
	openFile();
	readFile();
	closeFile();

	$INIT_RUN = 1;
}

sub openFile
{
	$DEBUG && abslog("opening DB");
	my $openstr = (-e $DB_FILE) ? "+<$DB_FILE" : ">$DB_FILE";
	$FH = FileHandle->new($openstr);
	defined($FH) || die "open on $DB_FILE failed";
}

sub closeFile
{
	$DEBUG && abslog("closing file.");
	$FH->close;
}

sub unlockFile
{
	#$DEBUG && print "unlock file.\n";
	$DEBUG && abslog("unlocking DB");
	flock($FH, LOCK_UN) || die "flock lock_un on $DB_FILE ($!)";
}

sub lockFile
{
	#$DEBUG && print "lock file.\n";
	$DEBUG && abslog("locking DB");
	flock($FH, LOCK_EX) || die "flock lock_ex on $DB_FILE ($!)";
}

sub writeFile
{
	$DEBUG && abslog("seek to beginning of DB");

	backupDB();

	#$DEBUG && print "write file.\n";
	$DEBUG && abslog("write file.");
	seek($FH, 0, 0) || confess "seek";
	truncate($FH, 0) || croak "truncate";

	$DEBUG && abslog("start writing DB");

	# write groups first
	print $FH "GROUPS:\n";
	foreach my $group_id (keys(%GROUPS)) {
		#if ($GROUPS{$group_id}->{name} eq 'default') { next; }
		print $FH "START\n";
		foreach my $key (keys(%{$GROUPS{$group_id}})) {
			print $FH "\t$key: $GROUPS{$group_id}->{$key}\n";
		}
		print $FH "END\n";
	}


	# write people
	print $FH "PEOPLE:\n";
	foreach my $person_id (keys(%PEOPLE)) {
		print $FH "START\n";
		foreach my $key (keys(%{$PEOPLE{$person_id}})) {
			print $FH "\t$key: $PEOPLE{$person_id}->{$key}\n";
		}
		print $FH "END\n";
	}

	my $val;
	print $FH "\nRESERVATIONS:\n";
	foreach my $res_id (keys(%RESERVATIONS)) {
		print $FH "START\n";
		foreach my $key (@RES_KEYS) {
			if (exists($RESERVATIONS{$res_id}->{$key})) {
				if ($key =~ /^(start|end)$/) {
					$val = join('.', @{$RESERVATIONS{$res_id}->{$key}});
					print $FH "\t$key: $val\n";
				} else {
					print $FH "\t$key: $RESERVATIONS{$res_id}->{$key}\n";
				}
			}
		}
		print $FH "END\n";
	}
	$DEBUG && abslog("done writing DB");
}

sub readFile
{
	my $mode = '';
	my $in_block = 0;
	my ($key, $val);
	my $curr;

	#$DEBUG && print "read file.\n";
	$DEBUG && abslog("read file.");
	seek($FH, 0, 0) || die "seek";

	$DEBUG && abslog("reading DB");

	%GROUPS = ();
	%PEOPLE = ();
	%RESERVATIONS = ();

	while(<$FH>) {
		chomp;
		($DEBUG > 2) && print "LINE: $_\n";
		if (/^\s*#/) { next; }		# skip WS
		if (/^PEOPLE:/) {
			$mode = 'people';
			next;
		} elsif (/^RESERVATIONS:/) {
			$mode = 'reservations';
			($DEBUG > 2) && print "found RESERVATIONS:\n";
			next;
		} elsif (/^GROUPS:/) {
			$mode = 'groups';
			($DEBUG > 2) && print "found GROUPS:\n";
			next;
		}
	
		if ($mode eq 'people') {
			if ($in_block) {
				if (/^END/) {
					$in_block = 0;
					if (!exists($curr->{id})) {
						die "no id found for person [$curr->{name}]";
					}
					if (exists($PEOPLE{$curr->{id}})) {
						die "person ID [$curr->{id}] occurs twice";
					}
					if (!exists($curr->{group})) {
						$curr->{group} = 1;
					}
					$PEOPLE{$curr->{id}} = $curr;
					undef $curr;
					next;
				}
				if (/^\s*(\S+):\s*(\S+(\s+\S+)*)\s*$/) {
					($key, $val) = ($1, $2);
					if (exists($PERSON_KEYS{$key})) {
						$curr->{$key} = $val;
					}
				}
			} elsif (/^START/) {
				$in_block = 1;
			}
		} elsif ($mode eq 'reservations') {
			if ($in_block) {
				if (/^END/) {
					$in_block = 0;
					($DEBUG > 2) && print "END: id = $curr->{id}\n";
					if (!exists($curr->{id})) {
						die "no id found for reservation [$curr->{start}]";
					}
					if (exists($RESERVATIONS{$curr->{id}})) {
						die "reservation ID [$curr->{id}] occurs twice";
					}
					$curr->{sjd} = AbsenceDate::julianDay(@{$curr->{start}});
					$curr->{ejd} = AbsenceDate::julianDay(@{$curr->{end}});
					$RESERVATIONS{$curr->{id}} = $curr;
					undef $curr;
					next;
				}
				if (/^\s*(\S+):\s*(\S+(\s+\S+)*)\s*$/) {
					($key, $val) = ($1, $2);
					($DEBUG > 2) && print "key = [$key], val = [$val]\n";
					if (exists($RESERVATION_KEYS{$key})) {
						($DEBUG > 2) && print "setting \$curr->{$key} = [$val]\n";
						if ($key =~ /^(start|end)$/) {
							$curr->{$key} = [split(/\./,$val)];
						} else {
							$curr->{$key} = $val;
						}
					}
				}
			} elsif (/^START/) {
				$in_block = 1;
				next;
			}
		} elsif ($mode eq 'groups') {
			if ($in_block) {
				if (/^END/) {
					$in_block = 0;
					($DEBUG > 2) && print "END: id = $curr->{id}\n";
					if (!exists($curr->{id})) {
						die "no id found for group [$curr->{name}]";
					}
					if (exists($GROUPS{$curr->{id}})) {
						die "group ID [$curr->{id}] occurs twice";
					}
					$GROUPS{$curr->{id}} = $curr;
					$GROUP_COUNT++;
					undef $curr;
					next;
				}
				if (/^\s*(\S+):\s*(\S+(\s+\S+)*)\s*$/) {
					($key, $val) = ($1, $2);
					($DEBUG > 2) && print "key = [$key], val = [$val]\n";
					if (exists($GROUP_KEYS{$key})) {
						($DEBUG > 2) && print "setting \$curr->{$key} = [$val]\n";
						$curr->{$key} = $val;
					}
				}
			} elsif (/^START/) {
				$in_block = 1;
				next;
			}
		}
	}
	if (!%GROUPS) {
		%GROUPS = ( 1 => { id => 1, name => 'default' } );
	}
}

sub logMsg
{
	my $msg = shift;
	
	my $fh = FileHandle->new(">>$LOGFILE");
	defined($fh) || die "open $LOGFILE for append";
	print $fh "".timestamp()." $msg\n";
	$fh->close;
}

sub timestamp
{
	my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
	my ($sec,$min,$hour,$mday,$mon,$year,@junk) = localtime(time);

	return sprintf("%d-%s-%d %02d:%02d",
		$mday, $months[$mon], $year+1900, $hour, $min);
}

sub backupDB
{
	# if current DB has been mulched, i.e., has zero length, don't backup
	my $size = (stat($DB_FILE))[7];
	if ($size == 0) { return; }

	my $i = AbsenceConfig::fetch('backup_copies');
	$DEBUG && abslog("backupDB: i=[$i]");

	my ($j, $old, $new);

	# first rotate old copies
	while($i > 1) {
		$j = $i - 1;
		$old = "${DB_FILE}.sav$j";
		$new = "${DB_FILE}.sav$i";
		$DEBUG && abslog("backupDB: looking for [$old]");
		if (-e $old) {
			rename($old, $new) || abslog("rename failed: $old --> $new");
		}
		$i--;
	}

	# new copy the database
	$new = "${DB_FILE}.sav1";
	if (system("cp $DB_FILE $new")) {
		abslog("'copy $DB_FILE $new' failed");
	}
}

1;

__END__

=head1 NAME

AbsenceDB.pm - interface to absence database.

=head1 SYNOPSIS

  use AbsenceDB;


Operations on people:

  @people_ids = getPeople($group_id);
  
  $name = getPerson($person_id);
	or
  ($name, $email, $group_id) = getPerson($person_id);

  $group_id = getGroup($person_id);

  $result = addPerson($person_id, $name, $email, $group_id);

  $result = deletePerson($person_id);


Operations on reservations:

  @reservations = getReservations($person_id);

  @reservations = getMonthReservations($person_id, $month, $year);

  $reservation_structure = getReservation($reservation_id);

  $result = addReservation($reservation_id, $start, $end, $type, $description);

  $result = deleteReservation($reservation_id);


Operations on groups:

  $result = addGroup($group_id, $group_name, $group_password);

  $result = deleteGroup($group_id);

  @groups = getGroups();

  $group_name = getGroupInfo($group_id);
	or
  ($group_name, $group_password) = getGroupInfo($group_id);

  $number_of_groups = groupCount();

  $mod_time = getModificationTime($group_id, $month, $year);

=head1 DESCRIPTION

C<AbsenceDB.pm> is a crude interface to a database.  In theory, the
underlying database could be anything.  You have perhaps noticed that
elsewhere I refer to B<absences>, but here they are called
B<reservations>?  That came about as a result of my original intention
to keep the system generic, i.e., not necessarily only for people.
Along the way I apparently got a little confused.

=head1 PERSON FUNCTIONS

=over 4

=item getPeople()

returns a list of person_ids belonging to the specified group.

=item getPerson()

retrieves information about the specified person_id.  In a scalar
context returns the name, in a list context returns (name, email).

=item addPerson()

creates a new person or modifies an existing one.  Parameters are
C<($person_id, $name, $email, $group_id)>.  If B<$person_id> equals
'new', a new person is created, otherwise B<$person_id> must
contain a valid person-id.  If an attempt is made to create a person
with the same name as an existing one, the operation fails and
the return-value is 'duplicate'.  On success returns 'ok'.

=item deletePerson()

guess what it does? Takes as single param a person-id.  Returns
'ok' on success, otherwise 'bad-id' if the person-id was not valid.


=head1 RESERVATION FUNCTIONS

=item getReservations()

returns a list of all reservations belonging to the specified person_id.

=item getMonthReservations()

finds all reservations for a given person_id in a given month and year.
returns a list of array-refs containing reservation-info in the
form C<[$reservation_id, $start, $end, $type, $description]>.  C<$start>
and C<$end> are array-refs of the form C<[$day, $month, $year]>.

Alternatetively the data-structure could be represented so:

  [
    $reservation_id,
    [
      $start_day,
      $start_month,
      $start_year,
    ],
    [
      $end_day,
      $end_month,
      $end_year,
    ],
    $type,
    $description,
  ]
	
=item getReservation()

given a reservation-id it returns a reference to an array in the
same format as B<getMonthReservations()> if the reservation-id is
valid, otherwise returns nothing.

=item addReservation()

adds a new reservation, or modifies an existing one. Parameters
are C<($res_id, $person_id, $start, $end, $type, $desc)>. If
B<$res_id> equals 'new', a new reservation is created, otherwise
B<$res_id> must contain valid reservation-id.

B<$start> and B<$end> must contain dates in the form "D.M.Y".

Returns:

=over 4

=item badday

the start or end day is invalid (i.e., February 30)

=item impossible

the start-day is greater than the end-day.

=item conflict

there is a conflict with another reservation.

=item ok

success.

=back

=item deleteReservation()

deletes a reservation.  Takes as single parameter a reservation-id.

returns:

=over 4

=item internalerror

The specified reservation-id does not exist.

=item conflict

The reservation-id disappeared suddenly.  Presumably someone else
got there before you.

=item ok

success.

=back


=head1 GROUP FUNCTIONS

=item addGroup()

creates a new group or modifies an existing one.  parameters are
($group_id, $name, $password).  If $group_id equals 'new', a new
group is created, otherwise $group_id must contain a valid group-id.
If an attempt is made to create a group with the same name as an
existing one, returns 'duplicate'.  Returns 'ok' on success.

=item deleteGroup()

guess what this does?  Takes as single parameter the group-id.  Returns
'ok' on success.

=item getGroups()

returns a list of group-ids corresponding to all existing groups.

=item getGroupInfo()

retrieves information about a group.  Takes as single parameter
the group-id.  Return-value depends on context: in a scalar
context, returns group-name, in a list context returns 
a list consisting of B<($group_name, $group_password)>.

=item getGroup()

returns the group-id to which a person belongs.  Takes as single param
the person-id.

=item groupCount()

returns the number of defined groups.

=item getModificationTime()

retrieves modification time for a particular group for a particular
month, if one is known.  This allows C<absence.pl> to determine
if a month-image needs to be created anew.  Returns the time in
epoch-second format if available, otherwise undef.

=back

=head1 AUTHOR

Robert Urban <urban@tru64.org>

Copyright (C) 2003 Robert Urban

=cut
