package AbsenceDate;

use Time::Local;
use POSIX qw(strftime);

BEGIN {
	@WEEKDAYS	= qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday);
	$VERSION	= '1.1';
}

#----------------------------------------------------------------------
# daysInMonth()
#
# $monthnum has format 1-12
#----------------------------------------------------------------------

sub daysInMonth
{
	my ($monthnum, $year) = @_;

	$monthnum--;
	my @daysinmonth = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

	if ($monthnum == 1) {
		# February
		return isLeapYear($year) ? 29 : 28;
	} else {
		return $daysinmonth[$monthnum];
	}
}

sub weekDayFromDate
{
	my ($day, $month, $year, $parm) = @_;

	my $time = timelocal(0, 0, 0, $day, $month - 1, $year - 1900);

	my $wd = (localtime($time))[6];

	if ($parm eq '-name') {
		return $WEEKDAYS[$wd];
	}

	return $wd;
}

sub isLeapYear
{
	my $year = shift;

	return (($year % 4 == 0) xor ($year % 100 == 0) xor ($year % 400 == 0));
}

sub yearDay
{
	my ($day, $month, $year) = @_;

	my $time = timelocal(0, 0, 0, $day, $month - 1, $year - 1900);
	return (localtime($time))[7] + 1;
}

## sub yearDay
## {
## 	my ($day, $month, $year) = @_;
## 
## 	my $month_num = 1;
## 	my $total = 0;
## 	while($month_num < $month) {
## 		$total += daysInMonth($month_num);
## 		$month_num++;
## 	}
## 	$total += $day;
## 
## 	$total;
## }

#----------------------------------------------------------------------
# monthName()
#
# takes month number in range 1 - 12
#----------------------------------------------------------------------
sub monthName
{
	my ($month_num, $parm) = @_;

	$month_num--;

	my @months = qw(January February March April May June
		July August Sepember October November December);

	if ($parm eq '-short') {
		return substr($months[$month_num], 0, 3);
	}

	return $months[$month_num];
}

#----------------------------------------------------------------------
#----------------------------------------------------------------------
#----------------------------------------------------------------------

#----------------------------------------------------------------------
# julianDay()
#
# calculates the Julian Day, as described in the "Calendar FAQ"
# (http://www.pauahtun.org/CalendarFAQ/cal/calendar24.html, section 2.15.1)
#
#----------------------------------------------------------------------

sub julianDay
{
	my ($day, $month, $year) = @_;

	use integer;
	my ($a, $y, $m);
	$a = (14 - $month)/12;
	$y = $year + 4800 - $a;
	$m = $month + 12 * $a - 3;

	my $jd = $day + (153 * $m + 2)/5 + 365 * $y + $y/4
		- $y/100 + $y/400 - 32045;

	return $jd;
}

sub julianToGregorian
{
	my $jd = shift;

	my ($a, $b, $c, $d, $e, $m, $day, $month, $year);

	use integer;
	$a = $jd + 32044;
	$b = (4 * $a + 3) / 146097;
	$c = $a - (146097 * $b) / 4;
	$d = (4 * $c + 3)/1461;
	$e = $c - (1461 * $d) / 4;
	$m = (5 * $e + 2) / 153;

	$day = $e - (153 * $m + 2) / 5 + 1;
	$month = $m + 3 - 12 * ($m / 10);
	$year = 100 * $b + $d - 4800 + $m / 10;

	return ($day, $month, $year);
}

##################################
# Gregorian Date from Julian Day #
##################################
sub julianToGregorian2 {
	my $jd = shift;

	use integer;
	my ($d,$i,$j,$l,$m,$n,$y);
	$l = $jd + 68569;
	$n = ( 4 * $l ) / 146097;
	$l = $l - ( 146097 * $n + 3 ) / 4;
	$i = ( 4000 * ( $l + 1 ) ) / 1461001;
	$l = $l - ( 1461 * $i ) / 4 + 31;
	$j = ( 80 * $l ) / 2447;
	$d = $l - ( 2447 * $j ) / 80;
	$l = $j / 11;
	$m = $j + 2 - ( 12 * $l );
	$y = 100 * ( $n - 49 ) + $i + $l;
	return ($d, $m, $y);
};
#----------------------------------------------------------------------
# calendarWeek()
#
# calculates the calendar week, as described in the "Calendar FAQ"
# (http://www.pauahtun.org/CalendarFAQ/cal/calendar24.html, section 5.8)
#
# returns an array consisting of calendar-week and year, where year
# is the year in which the calendar-week actually falls
#----------------------------------------------------------------------

sub calendarWeek
{
	my ($day, $month, $year) = @_;

	my $jd = julianDay($day, $month, $year);

	use integer;
	my($d1, $d4, $l);

	$d4 = ($jd + 31741 - ($jd % 7)) % 146097 % 36524 % 1461;
	$l = $d4/1460;
	$d1 = (($d4 - $l) % 365) + $l;

	my $cw = $d1/7 + 1;

	if (($month == 1) && ($cw > 51)) { $year--; }
	elsif (($month == 12) && ($cw == 1)) { $year++; }

	return ($cw, $year);
}

1;

__END__

=head1 NAME

AbsenceDate.pm - date functions needed by absence

=head1 SYNOPSIS

use AbsenceDate;

$julian_day = julianDay($day, $month, $year);
$days_in_month = daysInMonth($month_number, $year);
$week_day = weekDayFromDate($day, $month, $year, $option);
$month_name = monthName($month_number, $option);
$cw = calendarWeek($day, $month, $year);

=head1 DESCRIPTION

AbsenceDate.pm contains basic date-manipulation routines.

=head1 FUNCTIONS

=over 4

=item julianDay()

returns the "julian day", (days are numbered over many years).  This
allows simple comparions of dates without fumbling about with
days, months, and years.  Parameters are (day, month, year).

=item daysInMonth()

given a month and year, returns the number of days in that month.

=item weekDayFromDate();

returns the weekday-number for a particular date.  If C<$option>
is '-name', returns the name of the day instead of the number.

=item monthName()

given a month-number between 1 and 12, returns the name of the
month.  if C<$option> equals '-short', returns the first three
letters of the name.

=item calendarWeek()

returns the calendar-week in which a particular date falls.

=back

=head1 AUTHOR

Robert Urban <urban@tru64.org>

Copyright (C) 2003 Robert Urban

=cut
