package AbsenceTypes;

use FileHandle;

use AbsenceConfig;

BEGIN {
	$TYPE_FILE = AbsenceConfig::fetch('type_file');
	%DB = ();

	$INIT_RUN	= 0;
	$DEFAULT	= '';
	$VERSION 	= '1.1';
}

sub readFile
{
	my $fh = FileHandle->new($TYPE_FILE);
	defined($fh) || die "open $TYPE_FILE for reading";
	my ($type, $color);
	while(<$fh>) {
		if (/^\s*#/) { next; }
		if (/^\s*$/) { next; }
		if (/^Type:\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*(#.*)?$/) {
			$type = $1;
			$color = [$2, $3, $4];	# rgb
			$DB{$type} = $color;
		} elsif (/^\s*Default: (\S+)\s*$/) {
			$DEFAULT = $1;
		}
	}
	$fh->close;

	if ($DEFAULT && !exists($DB{$DEFAULT})) {
		$DEFAULT = '';
	}

	$INIT_RUN = 1;
}

sub getTypes
{
	if (!$INIT_RUN) { readFile(); }
	return keys(%DB);
}

sub getColor
{
	my $type = shift;

	if (!$INIT_RUN) { readFile(); }

	if (!exists($DB{$type})) {
		return undef;
	}

	return $DB{$type};
}

sub getHash
{
	if (!$INIT_RUN) { readFile(); }
	return \%DB;
}

sub getDefault
{
	if ($DEFAULT) {
		return $DEFAULT;
	}

	return undef;
}

sub modTime
{
	return (stat($TYPE_FILE))[9];
}

1;

__END__

=head1 NAME

AbsenceTypes.pm - primitive interface absence types

=head1 SYNOPSIS

  use AbsenceTypes;

  @types = AbsenceTypes::getTypes();
  $color_hash = AbsenceTypes::getColor('type-string');
  $href = AbsenceTypes::getHash();
  $default_type = AbsenceTypes::getDefault();
  $modification_time = AbsenceTypes::modTime();

=head1 DESCRIPTION

B<AbsenceTypes> provides a simple interface to a text file
containing type absence types (i.e., 'busy', 'vacation', etc).
The format of the file is:

  Type: <type-name1> Red Green Blue
  Type: <type-name2> Red Green Blue
  Type: <type-nameN> Red Green Blue
  Default: <type-name>

for example:

  Type: Vacation 255 00 00
  Type: Customer 00 255 00
  Type: Busy 90 90 255
  Type: Other 255 165 0
  
  Default: Customer

The <type-name> should not have any white-space in it. Red, Green,
and Blue are the corresponding color components.
Blank lines and lines beginning with hash ('#') are ignored.

=head1 FUNCTIONS

=over 4

=item B<getTypes()>

returns a list of type-names.

=item B<getColor()>

returns a reference to a list containing the red, green, and blue components.

=item B<getHash()>

returns a reference to a hash of the entire absence-type data-structure.
pretty much a hack.

=item B<getDefault()>

returns the default type-name.

=item B<modTime()>

returns the modification time of the 'database' file.  This allows scripts
to determine if images generated based on outdated information need to
be re-generated.

=head1 AUTHOR

Robert Urban <urban@tru64.org>

Copyright (C) Robert Urban 2003

=cut
