#!/usr/bin/perl

#-------------------------------------------------------------------------
# install.pl
#
# If you already have a previous version of absence installed, you
# can use install.pl to update your configuration.  It will not
# overwrite the datafiles (database-file, hoiday-db, absence-types-db,
# log-file).
#-------------------------------------------------------------------------

use FileHandle;
use File::Basename;
use Config;

BEGIN {
	# structure to hold information about parameters
	%PARMS = (
		data_dir_rel	=> ['ddr', 'relative path to data'],
		data_dir_abs	=> ['dda', 'absolute path to data',
							'$ENV{DOCUMENT_ROOT}', 'data_dir_rel'],
		cgi_dir_rel		=> ['cdr', 'relative path to cgi-bin', '/cgi-bin'],
		cgi_dir_abs		=> ['cda', 'absolute path to cgi-bin'],
		image_dir_rel	=> ['idr', 'relative path to image-dir',
							'data_dir_rel', 'img'],
		image_dir_abs	=> ['ida', 'absolute path to image-dir',
							'data_dir_abs', 'img'],
		database_file	=> ['dbf', 'absolute path to db-file',
							'data_dir_abs', 'absence.db'],
		modtime_db_file	=> ['mdbf', 'absolute path to modtime-file',
							'data_dir_abs', 'month_mod_times.db'],
		holiday_db_file	=> ['hdbf', 'absolute path to holiday-db',
							'data_dir_abs', 'holiday.db'],
		log_file		=> ['lf', 'absolute path to log-file',
							'data_dir_abs', 'log'],
		type_file		=> ['tf', 'absolute path to type-file',
							'data_dir_abs', 'absence-types.db'],
		main_script		=> ['ms', 'relative path to main-script',
							'cgi_dir_rel', 'absence.pl'],
		manage_script	=> ['mgs', 'relative path to manage-script',
							'cgi_dir_rel', 'absence-manage.pl'],
		control_script	=> ['cs', 'relative path to control-script',
							'cgi_dir_rel', 'absence-control.pl'],
		legend_style	=> ['ls', '"fit" or "constant:XX"', 'fit'],
		backup_copies	=> ['bc', 'number of db-backups', 10],
		image_type		=> ['it', '"png", "jpg", or "gif"', 'png'],
		map_type		=> ['mt', '"server" or "client"', 'client'],
		cookie_domain	=> ['cd', 'domain to use for cookies'],
	);

	@ORDER = qw(
		data_dir_rel
		data_dir_abs
		cgi_dir_rel
		cgi_dir_abs
		image_dir_rel
		image_dir_abs
		database_file
		modtime_db_file
		holiday_db_file
		log_file
		type_file
		main_script
		manage_script
		control_script
		legend_style
		backup_copies
		image_type
		map_type
		cookie_domain
	);

	@scripts = qw(
		absence.pl
		absence-click.pl
		absence-control.pl
		absence-manage.pl
		AbsenceConfig.pm
		AbsenceDate.pm
		AbsenceDB.pm
		AbsenceHolidays.pm
		AbsenceImage.pm
		AbsenceLog.pm
		AbsenceTypes.pm
	);

	@misc = qw(
		00README
		00README-INSTANCES
	);

	@html = qw(
		absence-help.html
	);

	#demo.html

	@db_files = qw(
		absence-types.db
		holiday.db
	);

	$VERBOSE = 1;
	$DEBUG = 0;
}

#----------------------------------------------------------------------
# main
#----------------------------------------------------------------------

# generate short-name to long-name param map
my ($long, $desc);
my $force = 0;

$WINDOZE = 0;
if ($Config{osname} eq 'MSWin32') {
	# blech, gasp, spit, hack, cough, wheeze, grunt, scream
	$WINDOZE = 1;
}

if (!$WINDOZE && ($> == 0)) {
	print "$0 should not be run as root.\n";
	print "please re-run as a normal user.\n";
	exit 1;
}

foreach (keys(%PARMS)) {
	($short, $desc) = @{$PARMS{$_}};
	$SHORTMAP{$short} = $_;
}

while($_ = shift) {
	s/^-//;
	if (exists($SHORTMAP{$_})) {
		$_ = $SHORTMAP{$_};
	}
	if (/^h$/) {
		usage();
		exit;
	} elsif (/^f/) {
		$force = 1;
	} elsif (/^dr/) {
		$ENV{DOCUMENT_ROOT} = shift;
	} elsif (exists($PARMS{$_})) {
		$val = shift;
		$tmp = ($val =~ /^def(ault)?$/) ? default($_) : $val;
		print "setting $_ to [$tmp]\n";
		$VALUES{$_} = $tmp;
	} elsif (exists($SHORTMAP{$_})) {
		$val = shift;
		$tmp = ($val =~ /^def(ault)?$/) ? default($_) : $val;
		print "setting $_ to [$tmp]\n";
		$VALUES{$SHORTMAP{$_}} = $tmp;
	}
}

print <<_EOF_;

Starting installation of absence V1.5...

I may need to ask you some configuration details.

The descriptions of the parameters refer to "absolute" and "relative"
paths.  By "absolute", I mean a normal path as seen by a script, for
example "/users/apache/html".  By "relative", I mean the path component
of a URL.  For example, given the URL

	http://www.spielwiese.de/apps/absence

The "relative path" in this case is "/apps/absence".

_EOF_

my $use_dr = 0;
foreach $p (@ORDER) {
	if (exists($VALUES{$p}) && $VALUES{$p}) {
		$VALS{$p} = expand($VALUES{$p});
	} else {
		$v = queryParam($p);
		$VALUES{$p} = $v;
		$VALS{$p} = expand($v);
		$DEBUG && print "got [$v]\n";
		$VALS{$p} =~ s/^\s+//;
	}
	if ($VALS{$p} =~ /\$ENV{DOCUMENT_ROOT}/) {
		$use_dr = 1;
		if ($WINDOZE) {
			print "WARNING: Windoze/IIS does not support DOCUMENT_ROOT\n";
		}
	}
}

my $dr;
if ($use_dr) {
	print "** USE DOCUMENT_ROOT **\n";
	if (!exists($ENV{DOCUMENT_ROOT})) {
		while(1) {
			$dr = query('enter the value of $DOCUMENT_ROOT');
			if (-d $dr) { last; }
			print "\n[$dr] is not a valid directory.\n";
		}
		$ENV{DOCUMENT_ROOT} = $dr;
	}
	foreach $p (@ORDER) {
		#print "[$p] VALS{$p} = [$VALS{$p}]\n";
		if ($VALS{$p} =~ /\$ENV\{DOCUMENT_ROOT\}/) {
			#$VALS{$p} =~ s/\$ENV{DOCUMENT_ROOT}/docRoot()/e;
			#$VALS{$p} =~ s/\$ENV{DOCUMENT_ROOT}/$ENV{DOCUMENT_ROOT}/e;
			$VALS{$p} = eval qq["$VALS{$p}"];
			$VALS{$p} =~ s!//!/!g;
		}
	}
}

if (($VALUES{map_type} ne 'client') && ($WINDOZE)) {
	print "A map_type of 'client' doesn't work on Windoze.\n";
	print "Automatically setting map_type to 'server'\n";
	$VALUES{map_type} = 'server';
}

#------------------------------------------------------------------
# I have collected all parameter values, so I can now start 
# creating directories (if necessary) and copying files
#------------------------------------------------------------------

print <<_EOF_;

I am about to perform the following steps:
- create [$VALS{cgi_dir_abs}] if necessary
- create [$VALS{data_dir_abs}] if necessary
- copy Perl modules and scripts to [$VALS{cgi_dir_abs}]
- copy HTML, README, and data files to [$VALS{data_dir_abs}]
- create a new top-level HTML entry point (index.html on unix)
- create a new empty database file in [$VALS{database_file}]
- create a new modtime file in [$VALS{modtime_db_file}]
- create a new log-file in [$VALS{log_file}]
- create a new image directory [$VALS{image_dir_abs}]
- set parameters in AbsenceConfig.pm
- change permissions on datafiles and directories

_EOF_

if (!yesNo('are you ready', 1)) {
	print "installation aborted by user.\n";
	exit;
}

if (! -e $VALS{cgi_dir_abs}) {
	verbose("directory [$VALS{cgi_dir_abs}] does not exist. creating...");
	makePath($VALS{cgi_dir_abs})
		|| die "mkdir for $VALS{cgi_dir_abs} failed. err=$CMDOUT\n";
}
if (! -e $VALS{data_dir_abs}) {
	verbose("directory [$VALS{data_dir_abs}] does not exist. creating...");
	makePath($VALS{data_dir_abs})
		|| die "mkdir for $VALS{data_dir_abs} failed. err=$CMDOUT\n";
}

#------------------------------------------------------------------
# figure out where source files should be
#------------------------------------------------------------------
my $dn = dirname($0);
print "\n";
while(1) {
	$SRCDIR = query('where is the unpacked kit ("." is OK)', $dn);
	if (! -e "$SRCDIR/$scripts[0]") {
		print "that doesn't seem to be right...\n";
	} else {
		last;
	}
}

#------------------------------------------------------------------
# copy files into place
#------------------------------------------------------------------
foreach $file (@scripts) {
	copyFile("$SRCDIR/$file", $VALS{cgi_dir_abs})
		|| die "copy of $file to $VALS{cgi_dir_abs} failed. err=$CMDOUT\n";
}

foreach $file (@misc, @html) {
	copyFile("$SRCDIR/$file", $VALS{data_dir_abs})
		|| die "copy of $file to $VALS{data_dir_abs} failed. err=$CMDOUT\n";
}

foreach $file (@db_files) {
	if (-e "$VALS{data_dir_abs}/$file") {
		verbose(" %% $file already present.  skip copy.");
		next;
	}
	copyFile("$SRCDIR/$file", $VALS{data_dir_abs})
		|| die "copy of $file to $VALS{data_dir_abs} failed. err=$CMDOUT\n";
}

#------------------------------------------------------------------
# create HTML entry point
#------------------------------------------------------------------
createIndex();

#------------------------------------------------------------------
# create empty database file
#------------------------------------------------------------------
touch($VALS{database_file}, 0660);
touch($VALS{modtime_db_file}, 0660);
touch($VALS{log_file}, 0660);
(-d $VALS{image_dir_abs}) ||
	myMkdir($VALS{image_dir_abs}, 0770) || die "mkdir failed";
editConfigFile("$VALS{cgi_dir_abs}/AbsenceConfig.pm", \%VALUES);

#------------------------------------------------------------------
# worry about file and directory permissions
#------------------------------------------------------------------
if ($WINDOZE) {
	#--------------------------------------------------------------
	# Windoze
	#--------------------------------------------------------------
	print <<'	_EOF_';

	Your platform is windows, so I'm going to let you worry about file
	and directory permissions.  I can offer these hints:
	- $VALS{cgi_dir_abs} should be executable for scripts.
	- $VALS{data_dir_abs} whould be writable.

	You will probably need to create two virtual directories, one called
	"absence", pointing to {$VALS{data_dir_abs}", the
	other called "cgi-bin" pointing to "$VALS{cgi_dir_abs}".

	Also, remember that multiple instances won't work on windoze.
	_EOF_
	exit;
	# god, it's good be done with windoze!
}

#--------------------------------------------------------------
# UNIX
#--------------------------------------------------------------
print <<_EOF_;

There are two ways of setting up file and directory permissions:
The quick-n-dirty way of the proper way.

The quick-n-dirty is not desirable because it causes various
files and directories to be set world-writable and world-executable.
This makes it insecure.  The Proper method uses group permissions
to allow the webserver process(es) access to the files in question.

Use the proper method if you can.

Quick-n-dirty:
  - set mode on data_dir_abs to		0777
  - set mode on image_dir_abs to	0777
  - set mode on database_file to	0666
  - set mode on modtime_db_file to	0666
  - set mode on log_file to		0666

Proper:
  - set group, mode on data_dir_abs to    <webserver>, 0770
  - set group, mode on image_dir_abs to   <webserver>, 0770
  - set group, mode on database_file to   <webserver>, 0660
  - set group, mode on modtime_db_file to <webserver>, 0660
  - set group, mode on log_file to 	  <webserver>, 0660

The catch to the proper way is that some steps must be
performed as root.  To that end this script will generate
a small bourne-shell script to be executed by root.

Please select the method you want to use.

_EOF_

my $root_script;
$ans = query('(q) quick-n-dirty / (p) proper', 'p', ['p', 'q']);
if (lc($ans) eq 'q') {
	quickPermissions();
	print "finished with installation\n";
} else {
	$root_script = properPermissions();
	print basename($0)." finished.  Please run '$root_script' as root\n";
}

print "\nThe author of absence is curious about who is using it,\n";
print "and would like to send a mail to himself via this script.\n";
print "answer no if you don't want to or don't have a direct connection\n";
print "to the internet.\n";

if (!$WINDOZE && yesNo('Send a mail to author', 1)) {
	sendMail();
}

if ($ans eq 'p') {
	my $str = "** Don't forget to run '$root_script' as root! **";
	my $len = length($str);
	print "\n", '*' x $len, "\n", $str, "\n", '*' x $len, "\n";
}

exit;

#======================================================================
# subroutines
#======================================================================

sub sendMail
{
	# only know how to send mail
	my @list = glob('/usr/*bin/sendmail');
	if (!@list) {
		print "can't find sendmail, giving up.\n";
		return;
	}

	print "I hope you don't mind answering a few questions...\n";
	print "The default answer is 'n/a'.  Feel free to leave the default.\n";
	my $purpose = query('What do you intend to do with absence', 'n/a');
	my $company = query('What does your company do', 'n/a');
	my $name = query('What is your name?', 'n/a');
	my $comments = query('Would you like to make any comments', 'none');

	my $fh = FileHandle->new("|$list[0] urban\@unix-beratung.de");
	print $fh <<_EOF_;
To: urban\@unix-beratung.de
Subject: someone using absence

purpose: $purpose
company does: $company
user: $name
comments: $comments
_EOF_

	print "Thanks!\n";
}

sub docRoot
{
	$ENV{DOCUMENT_ROOT};
}

sub expand
{
	my $rhs = shift;

	#if ($rhs =~ /\$/) {
	#	$rhs =~ s/\${([^}]+)}/substitute($1)/ge;
	#}

	while($rhs =~ /\${([^}]+)}/) {
		$rhs =~ s/\${([^}]+)}/substitute($1)/ge;
	}
	return $rhs;
}

sub substitute
{
	my $vname = shift;

	if (exists($VALUES{$vname})) {
		return $VALUES{$vname};
	} else {
		die "no var found";
	}

}

sub quickPermissions
{
	my %PERMS = (
		$VALS{data_dir_abs}		=> 0777,
		$VALS{image_dir_abs}	=> 0777,
		$VALS{database_file}	=> 0666,
		$VALS{modtime_db_file}	=> 0666,
		$VALS{log_file}			=> 0666,
	);

	foreach my $file (keys(%PERMS)) {
		my $p = sprintf('0%o', $PERMS{$file});
		verbose("setting [$file] to [$p]");
		$DEBUG || chmod($PERMS{$file}, $file);
	}
}

sub properPermissions
{
	my %PERMS = (
		$VALS{data_dir_abs}		=> 0770,
		$VALS{image_dir_abs}	=> 0770,
		$VALS{database_file}	=> 0660,
		$VALS{modtime_db_file}	=> 0660,
		$VALS{log_file}			=> 0660,
	);

	# first try to figure out if there is a group for the webserver
	my @grps = findWebGroup();
	my $grp;
	my $def;
	if (@grps) {
		print "Please tell me what under what group the webserver runs.\n";
		print "I found the following possibilities: ".join(', ', @grps)."\n";
		$def = $grps[0];
	} else {
		print "I was unable to find any groups that the webserver might use\n";
		print "Please tell me what under what group the webserver runs.\n";
	}

	while(1) {
		$grp = query('Which group shall I use', $def);
		if (getgrnam($grp)) { last; }
		print "\ngroup [$grp] does not exist.\n";
	}

	my $rs = "$SRCDIR/root-script";
	my $fh = FileHandle->new;
	open($fh, ">$rs") || die "open [$rs] for writing";
	print $fh "#!/bin/sh\n";
	foreach my $file (keys(%PERMS)) {
		$perms = $PERMS{$file};
		my $p = sprintf('0%o', $perms);
		verbose("(group,mode) for [$file] --> ($grp, $p)");
		print $fh "\nchgrp $grp $file\n";
		printf $fh "chmod 0%o $file\n", $perms;
	}
	$fh->close;
	chmod(0755, $rs) || die "chmod: $!";

	$rs;
}

# only used on U*X platforms
sub findWebGroup
{
	my $group;

	my @out;
	my $fh = FileHandle->new('/etc/group') || die 'open /etc/group to read';
	while(<$fh>) {
		$group = (split(':', $_))[0];
		if ($group =~ /(httpd|apach|web)/) {
			push(@out, $group);
		}
	}
	$fh->close;

	@out;
}

sub touch
{
	my ($file, $mode) = @_;

	if (-e $file) {
		verbose("touch: [$file] exists. do nothing.");
		return;
	}

	my $m = sprintf('0%o', $mode);
	verbose("touching [$file] with mode = ($m)");
	$DEBUG && return;

	my $fh = FileHandle->new;
	open($fh, ">$file") || die "touch on $file failed";
	$fh->close;
	chmod($mode, $file) || die "chmod: $!";
}

sub createIndex
{
	my $index_file = "$VALS{data_dir_abs}/"
		.($WINDOZE ? 'Default.htm' : 'index.html');

	if ($DEBUG) {
		$index_file = '/tmp/index.html';
		print "setting index file to /tmp/index.html\n";
	}

	my $fh = FileHandle->new;
	open($fh, ">$index_file") || die "open [$index_file] for writing";
	my $cgirel = $VALS{cgi_dir_rel};
	print $fh <<_EOF_;
<HTML>
<HEAD>
<TITLE>Absence</TITLE>
</HEAD>
<FRAMESET ROWS="60,*">
	<FRAME SRC="$cgirel/absence-control.pl" NAME=control FRAMEBORDER=0>
	<FRAME SRC="$cgirel/absence.pl" name=display>
</FRAMESET>
</HTML>
_EOF_
	$fh->close;
	chmod(0644, $index_file) || die "chmod: $!";
}

sub verbose
{
	my $str = shift;

	$VERBOSE && print "$str\n";
}

sub copyFile
{
	my ($src, $targ) = @_;

	my $cmd;
	if ($WINDOZE) {
		$src =~ s!/+!\\!g;
		$targ =~ s!/+!\\!g;
		$cmd = "copy $src $targ";
	} else {
		$cmd = "cp $src $targ";
	}

	myExec($cmd);
}

sub makePath
{
	my $path = shift;

	if (-d $path) { return 1; }

	my $cmd;
	if ($WINDOZE) {
		$path =~ s!/+!\\!g;
		$cmd = "mkdir $path";
	} else {
		$cmd = "mkdir -p $path";
	}

	myExec($cmd);
}

sub myMkdir
{
	my ($file, $perms) = @_;

	my $m = sprintf('0%o', $perms);
	verbose("CMD: mkdir($file, $m)");
	$DEBUG && return 1;

	return mkdir($file, $perms);
}

sub myExec
{
	my $cmd = shift;

	verbose("execing cmd [$cmd]");
	$DEBUG && return 1;

	$CMDOUT = `$cmd 2>&1`;

	#print "return value: [$?]\n";
	!$?;	# make success == 1
}

sub usage
{
	print "usage: $0 [options]\n  options:\n";
	my ($short, $long, $desc, $both);
	printf("  %-15s  %s\n", '-h', 'print help (this)');
	#printf("  %-15s  %s\n", '-f', "(f)orce. be bold.");
	printf("  %-15s  %s\n\n", '-dr <DOCROOT>', "set \$ENV{DOCUMENT_ROOT} to next arg");
	print "  configuration parameters:\n";
	printf("  %-5s %-16s %-27s %s\n",'short', 'long', 'description', 'default');
	printf("  %-5s %-16s %-27s %s\n",'-----', '----', '-----------', '-------');
	#printf("  %-5s %-16s %s\n",'short', 'long', 'description [default]');
	#printf("  %-5s %-16s %s\n",'-----', '----', '---------------------');
	foreach (@ORDER) {
		($short, $desc) = @{$PARMS{$_}};
		$desc =~ s/relative/rel./;
		$desc =~ s/absolute/abs./;
		$short = "-$short";
		$long = "-$_";
		$both = "$short / $long";
		printf("  %-5s %-16s %-27s %s\n", $short, $long, $desc, default($_));
		#printf("  %-5s %-16s %s\n", $short, $long, $desc.' ['.default($_).']');
	}
}

sub yesNo
{
	my ($prompt, $default) = @_;

	if ($default =~ /^1|0$/) {
		$default = ($default == 1) ? 'yes' : 'no';
	}
	my $ans = query($prompt, $default, ['yes', 'no', 'y', 'n']);
	if ($ans =~ /^y/i) {
		return 1;
	}
	return 0;
}
	
sub query
{
	my ($prompt, $default, $allowed) = @_;


	my $ans;
	my $def = ($default) ? " [$default]" : '';
	my $qs = "$prompt${def}? ";
	while(!$ans) {
		print "$qs";
		chomp($ans = <STDIN>);
		if (!length($ans) && $default) {
			$ans = $default;
		}
		if ($ans && ref($allowed) && !inList($ans, $allowed)) {
			print 'must be one of ('.join(',', @{$allowed}).")\n";
			$ans = '';
		}
	}

	$ans;
}

sub inList
{
	my ($thing, $lref) = @_;

	foreach (@{$lref}) {
		if ($thing eq $_) {
			return 1;
		}
	}

	0;
}

sub queryParam
{
	my $p = shift;

	my $default = default($p);
	my $desc = $PARMS{$p}->[1];
	if (($p eq 'data_dir_abs') || ($p eq 'cgi_dir_abs')) {
	  print "\n[You may want to use \$ENV{DOCUMENT_ROOT} in the path below]\n";
	}
	query($desc, $default);
}

sub query_param_old
{
	my $p = shift;

	my $default = default($p);
	my $ans;
	my $desc = $PARMS{$p}->[1];
	my $def = ($default) ? " [$default]" : '';
	my $qs = "$desc${def}? ";
	while(!$ans) {
		print "$qs";
		chomp($ans = <STDIN>);
		if (!length($ans) && $default) {
			$ans = $default;
		}
	}

	$ans;
}

sub default
{
	my $p = shift;

	my $default;
	my $path = 0;
	my @args = @{$PARMS{$p}};
	shift(@args);
	shift(@args);
	if (@args) {
		my @new;
		foreach (@args) {
			if (exists($PARMS{$_})) {
				$path = 1;
				push(@new, "\${$_}");
			} else {
				push(@new, $_);
			}
		}
		if ($path) {
			$default = join('/', @new);
		} else {
			$default = $args[0];
		}
	}

	$default;
}

sub editConfigFile
{
	my ($cf_path, $vref) = @_;

	$DEBUG && return;

	my $debug = 0;
	my %fixed;

	my $fh = FileHandle->new($cf_path) || die "file not found [$cf_path]\n";
	@lines = <$fh>;
	$fh->close;

	my @new;
	my $looking = 1;
	foreach (@lines) {
		$debug && print ">> $_";
		chomp;
		if ($looking) {
			if (/^\s*#/) { push(@new, $_); next; }
			if (/^(\s+\$?)(\S+)(\s*=>?\s*)(['"]?[^'";,]+['"]?)([;,]\s*(#.*)?)$/) {
				my ($pre, $varname, $mid, $val, $end) = ($1, $2, $3, $4, $5);
				$debug && print "  pre=[$pre] vn=[$varname] mid=[$mid] val=[$val] end=[$end]\n";
				if (exists($vref->{$varname}) && !exists($fixed{$varname})) {
					$debug && print "replacing [$val] with [$vref->{$varname}]\n";
					my $tmp = quote($vref->{$varname});
					push(@new, "$pre$varname$mid$tmp$end");
					$fixed{$varname} = 1;
				} else {
					push(@new, $_);
				}
			} else {
				push(@new, $_);
			}
		} else {
			push(@new, $_);
		}
		if (/cookie_domain/) { $looking = 0; }
	}

	my $dn = dirname($cf_path);
	my $tmpfile = "$dn/cf-$$";
	$fh = FileHandle->new;
	open($fh, ">$tmpfile") || die "open [$tmpfile] for writing";
	foreach (@new) {
		print $fh "$_\n";
	}
	$fh->close;
	chmod(0644, $tmpfile) || die "chmod: $!";

	if ($DEBUG) {
		print "leaving edited AbsenceConfig as [$tmpfile]\n";
		return;
	}

	rename($tmpfile, $cf_path) || die "rename failed";
}

sub quote
{
	my $val = shift;

	if ($val =~ /\$/) {
		return qq["$val"];
	} else {
		return qq['$val'];
	}
}
